/**************************************************************
 *                        Berliner Uhr
 *                      ================
 *                      Uwe Berger; 2009   
 * 
 * 
 * ---------
 * Have fun!
 * 
 ***************************************************************/ 
#include <avr/interrupt.h>
#include <util/delay.h>

#ifndef F_CPU
#define F_CPU 1000000UL     				
#endif


// Hardwarekonfiguration
// *********************
// Taster-Eingaenge
#define PORT_KEYX	PORTB
#define DDR_KEYX	DDRB
#define KEYX_PINX	PINB
#define KEY1_PIN	PB0
#define KEY2_PIN	PB1
#define KEY3_PIN	PB2

// Digit-Ausgaenge
#define PORT_DX		PORTC
#define DDR_DX		DDRC
#define D1			PC0
#define D2			PC1
#define D3			PC2

// Segment-Ausgaenge
#define PORT_SX		PORTD
#define DDR_SX		DDRD
#define S1			PD0
#define S2			PD1
#define S3			PD2
#define S4			PD3
#define S5			PD4
#define S6			PD5
#define S7			PD6
#define S8			PD7



// Timer Peroload-Wert
#define PRELOAD_T0 	152

// Tastenentprellung
#define DEBOUNCE_MS	100

// Tastenwerte
#define KEY1	1
#define KEY2	2
#define KEY3	3


uint8_t digit = 0;

volatile uint8_t click = 0;

typedef struct {uint8_t hs, ss, mm, hh, dd, mt, yy, wd, mez, mesz;} date_time_t;
volatile date_time_t dcf77, dt;

uint8_t old_mm= 60;

// Anzeigepuffer
volatile union btime_t {
	struct	{uint8_t d[3];} digit_data;
	struct	{
				unsigned m4:1;		//--> d[0]
				unsigned m3:1;
				unsigned m2:1;
				unsigned m1:1;
				unsigned m55:1;
				unsigned m50:1;
				unsigned m45:1;
				unsigned s:1;
				unsigned m40:1;		//--> d[1]
				unsigned m35:1;
				unsigned m30:1;
				unsigned m25:1;
				unsigned m20:1;
				unsigned m15:1;
				unsigned m10:1;
				unsigned m5:1;
				unsigned h4:1;		//--> d[2]
				unsigned h3:1;
				unsigned h2:1;
				unsigned h1:1;
				unsigned h20:1;
				unsigned h15:1;
				unsigned h10:1;
				unsigned h5:1;
			} d;
	} btime;


//************************************************************************
void my_delay_ms(uint16_t ms)
{
	uint16_t i;
	for (i=0; i<ms; i++) _delay_ms(1);
}


//************************************************************************
ISR(TIMER0_OVF_vect)
{
	TCNT0 = PRELOAD_T0;	

	// alles ausschalten	
	PORT_SX = 0xff;
	PORT_DX |= (1 << D1)|(1 << D2)|(1 << D3);

	// entsprechende Digit einschalten
	digit++;
	if (digit>2) digit = 0;	
	switch (digit) {
		case 0: PORT_DX  &= ~(1<<D1); break;
		case 1: PORT_DX  &= ~(1<<D2); break;
		case 2: PORT_DX  &= ~(1<<D3); break;
	}
	// Daten zum Digit ausgeben
	PORT_SX = ~btime.digit_data.d[digit];
	
}

//*********************************************************************
ISR(TIMER2_OVF_vect)
{
	// Preload setzen
	TCNT2     = 0;
	// ... mach was in der Hauptschleife
	click= 1;
}

//*********************************************************************
void add_time(void)
{
	dt.ss++;
	if (dt.ss > 59) {
		dt.ss = 0;
		dt.mm++;
		if (dt.mm > 59) {
			dt.mm = 0;
			dt.hh++;
			if (dt.hh > 23) dt.hh = 0;
		}	
	}
}

//*********************************************************************
void time2display(void)
{
	uint8_t i, hh;
	
	// Sekundenblinken
	if (dt.ss % 2) btime.d.s = 0; else btime.d.s = 1;
	// hat sich die Minute geaendert? Wenn nein, dann raus...
	if (old_mm == dt.mm) return;
	// Minute hat sich geaendert, also neues Anzeigebild berechnen	
	old_mm = dt.mm;
	// Anzeigepuffer loeschen
	btime.digit_data.d[0] = 0x00;
	btime.digit_data.d[1] = 0x00;
	btime.digit_data.d[2] = 0x00;
	// nochmal Sekundenblinken, weil Anzeigepuffer geloescht wurde
	if (dt.ss % 2) btime.d.s = 0; else btime.d.s = 1;
	// Sonderregelung: 00:00 = 24:00
	if ((dt.mm==0)&&(dt.hh==0)) hh=24; else hh=dt.hh;
	// 5er-Stunden
	for (i=1; i<5; i++) {
		if (hh/(i*5)) {
			switch (i) {
				case 1: btime.d.h5 = 1;  break;
				case 2: btime.d.h10 = 1; break;
				case 3: btime.d.h15 = 1; break;
				case 4: btime.d.h20 = 1; break;
			}
		}
	}
	// 1er-Stunden
	for (i=0; i<4; i++) {
		if (hh%5 > i) {
			switch (i) {
				case 0: btime.d.h1 = 1; break;
				case 1: btime.d.h2 = 1; break;
				case 2: btime.d.h3 = 1; break;
				case 3: btime.d.h4 = 1; break;
			}
		}
	}
	// 5er-Minuten
	for (i=1; i<12; i++) {
		if (dt.mm/(i*5)) {
			switch (i) {
				case 1:  btime.d.m5 = 1;  break;
				case 2:  btime.d.m10 = 1; break;
				case 3:  btime.d.m15 = 1; break;
				case 4:  btime.d.m20 = 1; break;
				case 5:  btime.d.m25 = 1; break;
				case 6:  btime.d.m30 = 1; break;
				case 7:  btime.d.m35 = 1; break;
				case 8:  btime.d.m40 = 1; break;
				case 9:  btime.d.m45 = 1; break;
				case 10: btime.d.m50 = 1; break;
				case 11: btime.d.m55 = 1; break;
			}
		}
	}
	// 1er-Minuten
	for (i=0; i<4; i++) {
		if (dt.mm%5 > i) {
			switch (i) {
				case 0: btime.d.m1 = 1; break;
				case 1: btime.d.m2 = 1; break;
				case 2: btime.d.m3 = 1; break;
				case 3: btime.d.m4 = 1; break;
			}
		}
	}
}

//************************************************************************
uint8_t scankey(volatile uint8_t *pinx, uint8_t pin) 
{
	if (bit_is_clear(*pinx, pin)) {
		my_delay_ms(DEBOUNCE_MS);
		if (bit_is_clear(*pinx, pin)) return 1;
	}
	return 0;
}

//************************************************************************
uint8_t getkey(void) 
{
	if (scankey(&KEYX_PINX, KEY1_PIN))	return KEY1;
	if (scankey(&KEYX_PINX, KEY2_PIN))	return KEY2;
	if (scankey(&KEYX_PINX, KEY3_PIN))	return KEY3;
	return 0;
}

//*********************************************************************
//*********************************************************************
//*********************************************************************
int main(void)
{
    
	// Eingaenge...
    PORT_KEYX &= ~(1 << KEY1_PIN);
    PORT_KEYX &= ~(1 << KEY2_PIN);
    PORT_KEYX &= ~(1 << KEY3_PIN);
    
    // Ausgaenge...
   	DDR_DX |= (1<<D1) | (1<<D2) | (1<<D3);
   	DDR_SX |= (1<<S1)|(1<<S2)|(1<<S3)|(1<<S4)|(1<<S5)|(1<<S6)|(1<<S7)|(1<<S8);
    
    // Timerinterrupt konfigurieren
    ASSR   = (1<< AS2);						// Timer2 asynchron takten
    TCNT2  = 0;
	TCCR2 |= (1 << CS22) | (1 << CS20);		// 128
	TIMSK |= (1 << TOIE2);
    
    TCNT0  = PRELOAD_T0;					// Preloadwert Timer0 setzen
	TCCR0 |= (1 << CS01)|(1 << CS00);		// 64
	TIMSK |= (1 << TOIE0);					// Timer0-OVF einschalten

	// ... und Interrupts einschalten
	sei();

	// Endlosschleife
	while(1){
		// ... es gibt was zu tun...
		if (click) {
			click = 0;
			// Uhrzeit hochzaehlen
			add_time();
			// Uhrzeit fuer Anzeige berechnen
			time2display();
		}
		// Tastaturabfrage
		switch (getkey()) {
			case KEY1:	{	// Stunden stellen
							dt.hh++;
							if (dt.hh > 23) dt.hh=0;
							old_mm = 60;
							time2display();
							break;
						}
			case KEY2:	{	// Minuten stellen
							dt.mm++;
							if (dt.mm > 59) dt.mm=0;
							time2display();
							break;
						}
			case KEY3:	{	// Sekunde auf Null stellen
							dt.ss=0;
							time2display();
							break;
						}
		}
	}
}
